/*
 * Copyright 2013-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.VerilogNode;
import ru.ispras.verilog.parser.walker.VerilogNodeVisitor;
import ru.ispras.verilog.parser.walker.VerilogTreeWalker;

/**
 * {@link VerilogSyntaxBackend} is a basic class for AST-level back-ends.
 * 
 * <p>The {@code start} method should be override in a subclass.</p>
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public class VerilogSyntaxBackend {
  /** Back-end name. */
  private final String name;
  /** User-defined visitor. */
  private final VerilogNodeVisitor visitor;

  /**
   * Creates an abstract back-end.
   * 
   * @param name the back-end name.
   */
  protected VerilogSyntaxBackend(final String name) {
    InvariantChecks.checkNotNull(name);

    this.name = name;
    this.visitor = null;
  }

  /**
   * Creates a visitor-based back-end.
   * 
   * @param name the back-end name.
   * @param visitor the user-defined visitor.
   */
  public VerilogSyntaxBackend(final String name, final VerilogNodeVisitor visitor) {
    InvariantChecks.checkNotNull(name);
    InvariantChecks.checkNotNull(visitor);

    this.name = name;
    this.visitor = visitor;
  }

  /**
   * Returns the name of the back-end.
   * 
   * @return the back-end name.
   */
  public final String getName() {
    return name;
  }

  /**
   * Processes the abstract syntax tree (AST).
   * 
   * @param root the AST's root.
   */
  public void start(final VerilogNode root) {
    InvariantChecks.checkNotNull(visitor);
    InvariantChecks.checkNotNull(root);

    final VerilogTreeWalker walker = new VerilogTreeWalker(root, visitor);
    walker.start();
  }
}
