/*
 * Copyright 2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.elaborator;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.util.InvariantChecks;

/**
 * {@link VerilogParameter} represents a parameter definition.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogParameter {

  public enum Type {
    /** Parameter definition. */
    PARAMETER(0),
    /** Parameter override (instantiation). */
    OVERRIDE(1),
    /** Parameter override (define parameter statement). */
    DEFPARAM(2),
    /** Local parameter (cannot be overridden). */
    LOCALPARAM(Integer.MAX_VALUE),
    /** Specification parameter (cannot be overridden). */
    SPECPARAM(Integer.MAX_VALUE);

    private final int priority;

    private Type(final int priority) {
      InvariantChecks.checkGreaterOrEqZero(priority);
      this.priority = priority;
    }

    public int getPriority() {
      return priority;
    }
  }

  private final Type type;
  private final String name;
  private final Data value;

  /**
   * Creates a parameter definition.
   * 
   * @param type the definition type.
   * @param name the parameter name.
   * @param value the parameter value.
   */
  public VerilogParameter(final Type type, final String name, final Data value) {
    InvariantChecks.checkNotNull(type);
    InvariantChecks.checkNotNull(name);
    InvariantChecks.checkNotNull(value);

    this.type = type;
    this.name = name;
    this.value = value;
  }

  public Type getType() {
    return type;
  }

  public String getName() {
    return name;
  }

  public Data getValue() {
    return value;
  }

  @Override
  public String toString() {
    return String.format("%s(%s=%s)", type, name, value);
  }
}
