/*
 * Copyright 2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogMinTypMax;
import ru.ispras.verilog.parser.model.basis.VerilogPath;

import java.util.EnumSet;

/**
 * {@code VerilogAssign} represents parameter definition statements ({@code defparam}).
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogDefineParameter extends VerilogNode {
  public static final Tag TAG = Tag.DEFINE_PARAMETER;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.noneOf(Tag.class);

  /** Path to the parameter. */
  private VerilogPath path;
  /** Constant expression. */
  private VerilogMinTypMax expression;

  /**
   * Constructs a parameter definition statement.
   * 
   * @param parent the parent node.
   */
  public VerilogDefineParameter(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.ELEMENT, parent);
  }

  /**
   * Constructs a copy of the parameter definition statement.
   * 
   * @param other the parameter definition statement to be copied.
   * @param parent the parent node.
   */
  public VerilogDefineParameter(final VerilogDefineParameter other, final VerilogNode parent) {
    super(other, parent);

    this.path = other.path != null ? other.path.clone() : null;
    this.expression = other.expression != null ? other.expression.clone() : null;
  }

  /**
   * Returns the parameter path.
   * 
   * @return the parameter path.
   */
  public VerilogPath getPath() {
    return path;
  }

  /**
   * Sets the parameter path.
   * 
   * @param path the path to be set.
   */
  public void setPath(final VerilogPath path) {
    InvariantChecks.checkNotNull(path);
    this.path = path;
  }

  /**
   * Returns the parameter expression.
   * 
   * @return the parameter expression.
   */
  public VerilogMinTypMax getExpression() {
    return expression;
  }

  /**
   * Sets the parameter expression.
   * 
   * @param expression the expression to be set.
   */
  public void setExpression(final VerilogMinTypMax expression) {
    InvariantChecks.checkNotNull(expression);
    this.expression = expression;
  }

  @Override
  public VerilogDefineParameter clone() {
    return new VerilogDefineParameter(this, getParentNode());
  }
}
