/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogDelay;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.model.basis.VerilogPathDescription;

import java.util.EnumSet;

/**
 * {@link VerilogPathDeclaration} represents path declarations.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogPathDeclaration extends VerilogNode {
  public static final Tag TAG = Tag.PATH_DECLARATION;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.of(Tag.PATH_DECLARATION);

  /** {@link Type} contains the path declaration types. */
  public enum Type {
    /** Simple or edge-sensitive path declartion: {@code description = delay}. */
    SIMPLE,
    /** If path declaration: {@code if(expression) ...}. */
    IF,
    /** If-none path declaration: {@code ifnone ...}. */
    IF_NONE
  }

  /** Type of the path declaration (by default, {@code SIMPLE}). */
  private Type type = Type.SIMPLE;
  /** Condition (only for {@code IF} path declarations). */
  private VerilogExpression expression;
  /** Path description (only for {@code SIMPLE} path declarations). */
  private VerilogPathDescription description;
  /** Path delay value (only for {@code SIMPLE} path declarations). */
  private VerilogDelay delay;

  /**
   * Creates a path declaration.
   * 
   * @param parent the parent node.
   */
  public VerilogPathDeclaration(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.OBJECT, parent);
  }

  /**
   * Creates a copy of path declaration.
   * 
   * @param other the path declaration to be copied.
   * @param parent the parent node.
   */
  public VerilogPathDeclaration(final VerilogPathDeclaration other, final VerilogNode parent) {
    super(other, parent);

    this.type = other.type;
    this.expression = other.expression != null ? other.expression.clone() : null;
    this.description = other.description != null ? other.description.clone() : null;
    this.delay = other.delay != null ? other.delay.clone() : null;
  }
  
  /**
   * Returns the type of the path declaration.
   * 
   * @return the path declaration type.
   */
  public Type getType() {
    return type;
  }

  /**
   * Sets the type of the path declaration.
   * 
   * @param type the type to be set.
   */
  public void setType(final Type type) {
    InvariantChecks.checkNotNull(type);
    this.type = type;
  }

  /**
   * Checks whether the path declaration is {@code IF}.
   * 
   * @return {@code true} iff the path declaration is {@code IF}.
   */
  public boolean isIf() {
    return type == Type.IF;
  }

  /**
   * Checks whether the path declaration is {@code IF_NONE}.
   * 
   * @return {@code true} iff the path declaration is {@code IF_NONE}.
   */
  public boolean isIfNone() {
    return type == Type.IF_NONE;
  }

  /**
   * Sets the path declaration type to {@code IF}.
   */
  public void setIf() {
    type = Type.IF;
  }

  /**
   * Sets the path declaration type to {@code IF_NONE}.
   */
  public void setIfNone() {
    type = Type.IF_NONE;
  }

  /**
   * Returns the condition of the path declaration (only for {@code IF}).
   * 
   * @return the condition.
   */
  public VerilogExpression getExpression() {
    return expression;
  }

  /**
   * Sets the condition of the path declaration (only for {@code IF}).
   * 
   * @param expression the condition to be set.
   */
  public void setExpression(final VerilogExpression expression) {
    InvariantChecks.checkNotNull(expression);
    this.expression = expression;
  }

  /**
   * Returns the path description (only for {@code SIMPLE}).
   * 
   * @return the path description.
   */
  public VerilogPathDescription getDescription() {
    return description;
  }

  /**
   * Sets the path description (only for {@code SIMPLE}).
   * 
   * @param description the description to be set.
   */
  public void setDescription(final VerilogPathDescription description) {
    InvariantChecks.checkNotNull(description);
    this.description = description;
  }

  /**
   * Returns the path delay value (only for {@code SIMPLE}).
   * 
   * @return the path delay value.
   */
  public VerilogDelay getDelay() {
    return delay;
  }

  /**
   * Sets the path delay value (only for {@code SIMPLE}).
   * 
   * @param delay the delay to be set.
   */
  public void setDelay(final VerilogDelay delay) {
    InvariantChecks.checkNotNull(delay);
    this.delay = delay;
  }

  /**
   * Returns the child path declaration (only for {@code IF} and {@code IF_NONE}).
   * 
   * @return the child path declaration.
   */
  public VerilogPathDeclaration getThen() {
    return (VerilogPathDeclaration) items(TAGS_CHILDREN).iterator().next();
  }

  /**
   * Sets the child path declaration (only for {@code IF}).
   * 
   * @param then the declaration to be set.
   */
  public void setThen(final VerilogPathDeclaration then) {
    add(then);
  }

  @Override
  public VerilogPathDeclaration clone() {
    return new VerilogPathDeclaration(this, getParentNode());
  }
}
