/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model.basis;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.Node.Kind;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.VerilogAttribute;
import ru.ispras.verilog.parser.model.VerilogProcedure;

import java.util.List;

/**
 * {@link VerilogExpression} represents expressions.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogExpression {
  /** The original expression (as it is presented in the source code). */
  private Node node;

  /** Path to the source-code function (if the expression is a function call). */
  private VerilogPath path;
  /** List of expression attributes (not currently in use). */
  private List<VerilogAttribute> attributes;
  /** Function declaration (for function calls only). */
  private VerilogProcedure declaration;

  /**
   * Creates a null expression.
   */
  public VerilogExpression() {}

  /**
   * Creates an expression with the given syntax.
   *
   * @param node the expression syntax.
   */
  public VerilogExpression(final Node node) {
    setNode(node);
  }

  /**
   * Creates a copy of the expression.
   * 
   * @param other the expression to be copied.
   */
  public VerilogExpression(final VerilogExpression other) {
    InvariantChecks.checkNotNull(other);

    this.node = other.node != null ? other.node.deepestCopy() : null;
    this.path = other.path != null ? other.path.clone() : null;

    this.node.setUserData(this);

    // There is no need to copy the attributes and the declaration.
    this.attributes = other.attributes;
    this.declaration = other.declaration;
  }

  /**
   * Returns the expression representation.
   *
   * @return the expression representation.
   */
  public Node getNode() {
    return node;
  }

  /**
   * Assigns this object's reference to the node's user data.
   *
   * @param node the node to be updated.
   */
  private void setUserData(final Node node) {
    if (node.getKind() == Kind.OPERATION) {
      // To access the expression attributes from the operation syntax.
      node.setUserData(this);
    }
  }

  /**
   * Sets the original representation of the expression and resets the elaborated representation.
   *
   * @param node the expression representation.
   */
  public void setNode(final Node node) {
    InvariantChecks.checkNotNull(node);

    setUserData(node);
    this.node = node;
  }

  /**
   * Checks whether the expression is null.
   *
   * @return {@code true} iff the expression is null.
   */
  public boolean isNull() {
    return node == null;
  }

  /**
   * Checks whether the expression is a value.
   *
   * @return {@code true} iff the expression is a value.
   */
  public boolean isValue() {
    return node.getKind() == Kind.VALUE;
  }

  /**
   * Checks whether the expression is a variable.
   *
   * @return {@code true} iff the expression is a variable.
   */
  public boolean isVariable() {
    return node.getKind() == Kind.VARIABLE;
  }

  /**
   * Checks whether the expression is an operation.
   *
   * @return {@code true} iff the expression is an operation.
   */
  public boolean isOperation() {
    return node.getKind() == Kind.OPERATION;
  }

  /**
   * Returns the value node (if the expression is a value).
   *
   * @return the value node.
   */
  public NodeValue getValue() {
    return (NodeValue) node;
  }

  /**
   * Returns the literal.
   *
   * @return the literal if the expression is a value.
   */
  public VerilogLiteral getLiteral() {
    final NodeValue value = getValue();
    final Data data = value.getData();

    return (VerilogLiteral) data.getUserData();
  }

  /**
   * Returns the variable node (if the expression is a variable).
   *
   * @return the variable node.
   */
  public NodeVariable getVariable() {
    return (NodeVariable) node;
  }

  /**
   * Returns the operation node (if the expression is an operation).
   *
   * @return the operation node.
   */
  public NodeOperation getOperation() {
    return (NodeOperation) node;
  }

  /**
   * Returns the path to the source-code function.
   *
   * @return the path to the function.
   */
  public VerilogPath getPath() {
    return path;
  }

  /**
   * Sets a path to the source-code function.
   *
   * @param path the path to be set.
   */
  public void setPath(final VerilogPath path) {
    InvariantChecks.checkNotNull(path);
    this.path = path;
  }

  /**
   * Returns the expression attributes.
   *
   * @return the attributes.
   */
  public List<VerilogAttribute> getAttributes() {
    return attributes;
  }

  /**
   * Sets the expression attributes.
   *
   * @param attributes the attributes to be set.
   */
  public void setAttributes(final List<VerilogAttribute> attributes) {
    InvariantChecks.checkNotNull(attributes);
    this.attributes = attributes;
  }

  /**
   * Returns the function declaration (for function calls only).
   *
   * @return the function declaration.
   */
  public VerilogProcedure getDeclaration() {
    return declaration;
  }

  /**
   * Sets the function declaration (for function calls only).
   *
   * @param declaration the function declaration to be set.
   */
  public void setDeclaration(final VerilogProcedure declaration) {
    InvariantChecks.checkNotNull(declaration);
    this.declaration = declaration;
  }

  @Override
  public String toString() {
    return getNode().toString();
  }

  @Override
  public VerilogExpression clone() {
    return new VerilogExpression(this);
  }
}
