/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model.basis;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.VerilogAttribute;

import java.util.List;

/**
 * {@link VerilogMinTypMax} represents the min:typ:max expressions.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogMinTypMax {
  /** Minimal value expression. */
  private VerilogExpression min;
  /** Typical value expression. */
  private VerilogExpression typ;
  /** Maximal value expression. */
  private VerilogExpression max;

  /** Expression attributes (not currently in use). */
  private List<VerilogAttribute> attributes;

  /**
   * Creates a min:typ:max expression.
   */
  public VerilogMinTypMax() {}

  /**
   * Creates a copy of the min:typ:max expression.
   * 
   * @param other the expression to be copied.
   */
  public VerilogMinTypMax(final VerilogMinTypMax other) {
    InvariantChecks.checkNotNull(other);

    this.min = other.min != null ? other.min.clone() : null;
    this.typ = other.typ != null ? other.typ.clone() : null;
    this.max = other.max != null ? other.max.clone() : null;

    // There is no need to copy attributes.
    this.attributes = other.attributes;
  }

  /**
   * Checks whether the expression is usual.
   * 
   * @return {@code true} iff the expression is usual.
   */
  public boolean isUsual() {
    return typ == null && max == null;
  }

  /**
   * Returns the minimal value expression.
   * 
   * @return the minimal value expression.
   */
  public VerilogExpression getMinExpression() {
    return min;
  }

  /**
   * Sets the minimal value expression.
   * 
   * @param min the expression to be set.
   */
  public void setMinExpression(final VerilogExpression min) {
    InvariantChecks.checkNotNull(min);
    this.min = min;
  }

  /**
   * Returns the typical value expression.
   * 
   * @return the typical value expression.
   */
  public VerilogExpression getTypExpression() {
    return typ;
  }

  /**
   * Sets the typical value expression.
   * 
   * @param typ the expression to be set.
   */
  public void setTypExpression(final VerilogExpression typ) {
    InvariantChecks.checkNotNull(typ);
    this.typ = typ;
  }

  /**
   * Set the minimal, the typical, and the maximal expressions.
   * 
   * @param other the min:typ:max expression to be set. 
   */
  public void setMinTypMaxExpression(final VerilogMinTypMax other) {
    InvariantChecks.checkNotNull(other);

    this.min = other.getMinExpression();
    this.typ = other.getTypExpression();
    this.max = other.getMaxExpression();
  }

  /**
   * Returns the maximal value expression.
   * 
   * @return the maximal value expression.
   */
  public VerilogExpression getMaxExpression() {
    return max;
  }

  /**
   * Sets the maximal value expression.
   * 
   * @param max the expression to be set.
   */
  public void setMaxExpression(final VerilogExpression max) {
    InvariantChecks.checkNotNull(max);
    this.max = max;
  }

  /**
   * Returns the expression attributes.
   * 
   * @return the attributes.
   */
  public List<VerilogAttribute> getAttributes() {
    return attributes;
  }

  /**
   * Sets the expression attributes.
   * 
   * @param attributes the attributes to be set.
   */
  public void setAttributes(final List<VerilogAttribute> attributes) {
    InvariantChecks.checkNotNull(attributes);
    this.attributes = attributes;
  }

  @Override
  public String toString() {
    return isUsual() ? min.toString() : String.format("%s:%s:%s", min, typ, max);
  }

  @Override
  public VerilogMinTypMax clone() {
    return new VerilogMinTypMax(this);
  }
}
