/*
 * Copyright 2016-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.verilog.parser.transformer;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.NodeTransformer;
import ru.ispras.fortress.transformer.Reducer;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.transformer.TransformerRule;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.VerilogLogger;
import ru.ispras.verilog.parser.interpreter.VerilogCalculator;
import ru.ispras.verilog.parser.model.VerilogNode;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

/**
 * {@link VerilogTransformerOperation} implements Verilog-specific expression transformation rules.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class VerilogTransformerOperation extends VerilogTransformer {

  /**
   * Reduces the expression and applies type casting
   *
   * @param node the AST node to be transformed.
   */
  public static void transform(final VerilogNode node) {
    InvariantChecks.checkNotNull(node);

    final VerilogTransformer transformer = new VerilogTransformerOperation();
    transformer.run(node);
  }

  private static NodeTransformer ruleSet = getRuleSet();

  @Override
  public Node transform(final Node node) {
    InvariantChecks.checkNotNull(node);

    final Node transformedNode = Transformer.transform(node, ruleSet);
    final Node reducedNode = VerilogCalculator.reduce(transformedNode);

    return Transformer.standardize(reducedNode);
  }

  private static NodeTransformer getRuleSet() {
    final NodeTransformer transformer = new NodeTransformer();

    // The following rule transforms condition of the if-then-else operation.
    final TransformerRule iteRule = new TransformerRule() {
      @Override
      public boolean isApplicable(final Node node) {
        return ExprUtils.isOperation(node, StandardOperation.ITE);
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation operation = (NodeOperation) node;
        InvariantChecks.checkTrue(operation.getOperandCount() == 3);

        /* Cast "if" operand for ITE operation to boolean type. */
        final Node nodeIf = operation.getOperand(0);
        final Node condition = coerce(nodeIf, DataType.BOOLEAN);

        /* Cast "then" and "else" operands for ITE operation to maximum sized type. */
        final Node nodeThen = operation.getOperand(1);
        final DataType thenType = nodeThen.getDataType();

        final Node nodeElse = operation.getOperand(2);
        final DataType elseType = nodeElse.getDataType();

        final DataType maxType = max(thenType, elseType);
        final Node coercedThen = coerce(nodeThen, maxType);
        final Node coercedElse = coerce(nodeElse, maxType);

        InvariantChecks.checkNotNull(condition);

        return Nodes.ite(condition, coercedThen, coercedElse);
      }
    };

    transformer.addRule(StandardOperation.ITE, iteRule);

    // The following rule transforms operands of the boolean operations.
    final TransformerRule boolRule = new TransformerRule() {
      @Override
      public boolean isApplicable(final Node node) {
        return ExprUtils.isOperation(node, StandardOperation.getBoolOperandOperations());
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation operation = (NodeOperation) node;
        final List<Node> operands = new LinkedList<>();

        boolean wrongTypeOperands = false;
        for (final Node operand : operation.getOperands()) {
          if (!ExprUtils.isType(DataType.BOOLEAN, operand)
              && !ExprUtils.isOperation(operand, StandardOperation.FUNCTION)) {
            final Node reducedOperand = VerilogCalculator.reduce(operand);
            final Node boolOperand = coerce(reducedOperand, DataType.BOOLEAN);
            InvariantChecks.checkNotNull(boolOperand,
                String.format("Coerced to NULL: %s", reducedOperand));

            operands.add(boolOperand);
            wrongTypeOperands = true;
          } else {
            operands.add(operand);
          }
        }

        return wrongTypeOperands
            ? new NodeOperation(operation.getOperationId(), operands)
            : node;
      }
    };

    for (final Enum<?> operation : StandardOperation.getBoolOperandOperations()) {
      transformer.addRule(operation, boolRule);
    }

    // The following rule transforms parameters and boolean operands of the bit vector operations.
    final TransformerRule bvRule = new TransformerRule() {
      @Override
      public boolean isApplicable(final Node node) {
        return ExprUtils.isOperation(node, StandardOperation.getBvOperandOperations());
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation operation = (NodeOperation) node;
        final Enum<?> opId = operation.getOperationId();

        final List<Node> operands = new ArrayList<>(operation.getOperandCount());

        boolean wrongTypeOperands = false;
        for (int i = 0; i < operation.getOperandCount(); i++) {
          final Node operand = operation.getOperand(i);
          InvariantChecks.checkNotNull(operand);

          if (i < StandardOperation.getParameterCount(opId)
              && !ExprUtils.isType(DataType.INTEGER, operand)
              && ExprUtils.getVariables(operand).isEmpty()) {
            final Node reducedOperand = VerilogCalculator.reduce(operand);
            final Node intOperand = coerce(reducedOperand, DataType.INTEGER);
            InvariantChecks.checkNotNull(intOperand, String.format("%s", reducedOperand));

            operands.add(intOperand);
            wrongTypeOperands = true;
          } else if (ExprUtils.isType(DataType.BOOLEAN, operand)
              && i >= StandardOperation.getParameterCount(opId)) {
            final Node reducedOperand = VerilogCalculator.reduce(operand);
            final Node bvOperand = coerce(reducedOperand, DataType.bitVector(1));
            InvariantChecks.checkNotNull(bvOperand, String.format("%s", reducedOperand));

            operands.add(bvOperand);
            wrongTypeOperands = true;
          } else {
            operands.add(operand);
          }
        }

        return wrongTypeOperands ? new NodeOperation(opId, operands) : node;
      }
    };

    for (final Enum<?> operation : StandardOperation.getBvOperandOperations()) {
      transformer.addRule(operation, bvRule);
    }

    // The following rule transforms parameters and boolean operands of the bit vector operations.
    final TransformerRule bvextractRule = new TransformerRule() {
      @Override
      public boolean isApplicable(final Node node) {
        return ExprUtils.isOperation(node, StandardOperation.BVEXTRACT);
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation operation = (NodeOperation) node;

        final Node operand0 = operation.getOperand(0);
        final Node operand1 = operation.getOperand(1);
        final Node operand2 = operation.getOperand(2);

        if (ExprUtils.isValue(operand0) && ExprUtils.isValue(operand1)) {
          final NodeValue value0 = (NodeValue) operand0;
          final NodeValue value1 = (NodeValue) operand1;

          final int intValue0 = value0.getInteger().intValue();
          final int intValue1 = value1.getInteger().intValue();

          if (intValue0 < intValue1) {
            return Nodes.bvextract(operand1, operand0, operand2);
          }
        }

        return Nodes.bvextract(operand0, operand1, operand2);
      }
    };

    transformer.addRule(StandardOperation.BVEXTRACT, bvextractRule);

    // The following rule transforms operands of "same type operand" operations.
    final TransformerRule sameTypeRule = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {
        return ExprUtils.isOperation(node, StandardOperation.getSameOperandOperations());
      }

      @Override
      public Node apply(final Node node) {

        final NodeOperation operation = (NodeOperation)node;

        if (operation.getOperandCount() != 2) {
          return node;
        }

        /* Cast operands to maximum size type. */
        final Node firstOp = operation.getOperand(0);
        final Node secondOp = operation.getOperand(1);

        final DataType firstOpType = firstOp.getDataType();
        final DataType secondOpType = secondOp.getDataType();
        final DataType maxType = max(firstOpType, secondOpType);

        final Node first = coerce(firstOp, maxType);
        final Node second = coerce(secondOp, maxType);

        return new NodeOperation(operation.getOperationId(), first, second);
      }
    };

    for (final Enum<?> operation : StandardOperation.getSameOperandOperations()) {
      transformer.addRule(operation, sameTypeRule);
    }

    // The following rule transforms operands of "integer" operations.
    final TransformerRule intRule = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {
        return ExprUtils.isOperation(node, StandardOperation.getIntOperandOperations());
      }

      @Override
      public Node apply(final Node node) {

        final NodeOperation operation = (NodeOperation) node;

        final List<Node> intOperands = new LinkedList<>();

        for (int i = 0; i < operation.getOperandCount(); i++) {

          intOperands.add(coerce(operation.getOperand(i), DataType.INTEGER));
        }

        return new NodeOperation(operation.getOperationId(), intOperands);
      }
    };

    for (final Enum<?> operation : StandardOperation.getIntOperandOperations()) {
      transformer.addRule(operation, intRule);
    }

    return transformer;
  }

  private static DataType max(final DataType first, final DataType second) {

    return first.getSize() > second.getSize() ? first : second;
  }

  private static Node coerce(final Node node, final DataType type) {

    final Node coerced = TypeConversion.coerce(Reducer.reduce(node), type);

    if (coerced == null) {
      VerilogLogger.warn("Can't cast %s to %s type; no changes.", node, type);
      return node;
    }

    return coerced;
  }
}
