/*
 * Copyright 2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.transformer;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.NodeProvider;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.elaborator.VerilogContext;
import ru.ispras.verilog.parser.elaborator.VerilogDescriptor;
import ru.ispras.verilog.parser.elaborator.VerilogParameter;
import ru.ispras.verilog.parser.model.VerilogNode;

import java.util.Map;

/**
 * {@link VerilogTransformerVariableSubstitute} implements an engine that substitutes variables with expressions.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogTransformerVariableSubstitute extends VerilogTransformer {

  public static void transform(
      final VerilogNode node,
      final Map<String, VerilogDescriptor> variables,
      final VerilogContext context,
      final Map<String, VerilogParameter> parameters) {
    InvariantChecks.checkNotNull(node);
    InvariantChecks.checkNotNull(context);
    InvariantChecks.checkNotNull(parameters);

    final VerilogTransformer transformer =
        new VerilogTransformerVariableSubstitute(variables, context, parameters);
    transformer.run(node);
  }

  /** All elaborated variables. */
  private final Map<String, VerilogDescriptor> variables;
  /** Contains bindings to be applied. */
  private final VerilogContext context;
  /** Contains parameters to be applied. */
  private final Map<String, VerilogParameter> parameters;

  public VerilogTransformerVariableSubstitute(
      final Map<String, VerilogDescriptor> variables,
      final VerilogContext context,
      final Map<String, VerilogParameter> parameters) {
    InvariantChecks.checkNotNull(context);
    InvariantChecks.checkNotNull(parameters);

    this.variables = variables;
    this.context = context;
    this.parameters = parameters;
  }

  @Override
  public Node transform(final Node node) {
    final Map<String, Node> bindings = context.getBindings();

    final Node result = Transformer.substitute(node, new NodeProvider() {
      @Override
      public Node getNode(final Variable variable) {
        final Node newNode;

        final String localName = variable.getName();
        final Node binding = bindings.get(localName);

        if (binding != null) {
          // Apply the binding.
          newNode = binding;
        } else {
          final String fullName = context.getFullName(localName);
          final VerilogDescriptor descriptor = variables.get(fullName);

          if (descriptor != null) {
            // Access is done via a qualified name not presented in the bindings.
            newNode = descriptor.getVariable();
          } else {
            // Variable is not a real one (a generate variable or a module reference).
            newNode = new NodeVariable(variable);
          }
        }

        // Replace parameters and generate variables with the corresponding values.
        return Transformer.substitute(newNode, new NodeProvider() {
          @Override
          public Node getNode(final Variable variable) {
            final String fullName = variable.getName();

            final VerilogParameter parameter = parameters.get(fullName);
            if (parameter != null) {
              // Substitute the parameter with its value.
              return new NodeValue(parameter.getValue());
            }

            final Data value = context.getGenvarValue(fullName);
            if (value != null) {
              // Substitute the generate variable with its value.
              return new NodeValue(value);
            }

            final VerilogDescriptor descriptor = variables.get(fullName);
            if (descriptor != null) {
              // Replace the variable with the elaborated one.
              return descriptor.getVariable();
            }

            // Variable is a module reference.
            return new NodeVariable(variable);
          }
        });
      }
    });

    return result;
  }
}
